# -*- coding: utf-8 -*-
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Sergi Blanch-Torne"
__email__ = "sblanch@cells.es"
__copyright__ = "Copyright 2015, CELLS / ALBA Synchrotron"
__license__ = "GPLv3+"


'''
    This file contains the necessary object to define the tree structure
    of the scpi commands. From the root to the latest nodes before the leaves
    they are Component objects that are subclass of dict. the leaves are
    an special type of components that have also the read (and the write
    optional) for the actions.
'''


try:
    from .logger import Logger as _Logger
except:
    from logger import Logger as _Logger
try:
    from numpy import ndarray as _np_ndarray
    from numpy import float16 as _np_float16
    from numpy import float32 as _np_float32
    from numpy import float64 as _np_float64
    from numpy import float128 as _np_float128
    _np = True
except:
    _np = False
try:
    from scipy import ndarray as _sp_ndarray
    from scipy import float16 as _sp_float16
    from scipy import float32 as _sp_float32
    from scipy import float64 as _sp_float64
    from scipy import float128 as _sp_float128
    _sp = True
except:
    _sp = False


if _np and not _sp:
    _float16 = _np_float16
    _float32 = _np_float32
    _float64 = _np_float64
    _float128 = _np_float128
elif _sp:
    _float16 = _sp_float16
    _float32 = _sp_float32
    _float64 = _sp_float64
    _float128 = _sp_float128


MINIMUMKEYLENGHT = 4
CHNUMSIZE = 2


class DictKey(_Logger, str):
    '''
        This class is made to allow the dictionary keys to find a match using
        the shorter strings allowed in the scpi specs.
    '''
    def __init__(self, value, *args, **kargs):
        value = "%s" % value
        super(DictKey, self).__init__(*args, **kargs)
        if not value.isalpha():
            raise NameError("key shall be strictly alphabetic")
        self._name = value
        if 0 < len(self._name) < MINIMUMKEYLENGHT:
            self.minimum = len(value)
        else:
            self.minimum = MINIMUMKEYLENGHT
        if len(self._name) < self.minimum:
            raise NameError("value string shall be almost "
                            "the minimum size")

    @property
    def minimum(self):
        return self._minimum

    @minimum.setter
    def minimum(self, value):
        if type(value) != int:
            raise TypeError("minimum shall be an integer")
        self._minimum = value

    def __str__(self):
        return self._name

    def __repr__(self):
        return "%s%s" % (self._name[0:self._minimum].upper(),
                         self._name[self._minimum:])

    def __eq__(self, other):  # => self == other
        '''
            Compare if those two names matches reducing the name until the
            minimum size.
        '''
        if type(other) == DictKey:
            otherName = other._name.lower()
        elif type(other) == str:
            otherName = other.lower()
        else:
            otherName = ''
        selfName = self._name.lower()
        self._debug("Comparing %s to %s" % (selfName, otherName))
        while len(selfName) >= len(otherName) and \
                len(selfName) >= self._minimum:
            if selfName == otherName:
                self._debug("Found match! %s == %s" % (selfName, otherName))
                return True
            if len(selfName) > self._minimum:
                self._debug("No match found, reducing %s to %s"
                            % (selfName, selfName[:-1]))
            selfName = selfName[:-1]
        return False

    def __ne__(self, other):  # => self != other
        return not self == other

    def is_(self, other):  # => self is other
        return self == other

    def is_not(self, other):  # => not self is other
        return self is not other


class Attribute(DictKey):
    '''
        Leaf node of the scpi command tree
        TODO: explain property_cb difference with {read,write}_cb

        Example: COMPonent:COMPonent:ATTRibute
    '''
    def __init__(self, *args, **kargs):
        super(Attribute, self).__init__(*args, **kargs)
        self._parent = None
        self._read_cb = None
        self._write_cb = None
        self._hasChannels = False
        self._channelTree = None
        self._allowedArgins = None
        self._debug("Build a Attribute object %s" % (self.name))

    def __str__(self):
        fullName = "%s" % (self._name)
        parent = self._parent
        while parent is not None and parent._name is not None:
            fullName = "".join("%s:%s" % (parent._name, fullName))
            parent = parent._parent
        return fullName

    def __repr__(self):
        indentation = "\t"*self.depth
        return ""  # .join("\n%s%s"%(indentation,DictKey.__repr__(self)))

    @property
    def name(self):
        return self._name

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        if value is not None:
            self.logEnable(value.logState())
            self.logLevel(value.logGetLevel())
            self._parent = value

    @property
    def hasChannels(self):
        return self._hasChannels

    @property
    def allowedArgins(self):
        return self._allowedArgins

    @allowedArgins.setter
    def allowedArgins(self, value):
        # TODO: also than an specific set of values, provide a way to restrict
        #       argins in a bounded region. Like a range (or ranges) of values.
        if value is not None and type(value) is not list:
            raise TypeError("Allowed argins expects a list")
        self._allowedArgins = []
        for element in value:
            self._allowedArgins.append(str(element))
        # TODO: those strings shall also follow the key length feature
        #       if the length string is at least the minimum.
        #       This is like 'FALSe' == 'FALS' in scpi
        #       replace 'str' by 'DictKey'

    def checkChannels(self):
        if self.parent is not None and self.parent.hasChannels:
            self._hasChannels = True
            self._channelTree = self._getChannels()
            self._debug("%s: Channels found for %s component: %s"
                        % (self.name, self.parent.name,
                           ["%s" % x.name for x in self._channelTree]))
        else:
            self._debug("%s: No channels found" % (self.name))

    def _getChannels(self):
        if self.parent is not None and self.parent.hasChannels:
            return self.parent._getChannels()
        return None

    @property
    def read_cb(self):
        return self._read_cb

    @read_cb.setter
    def read_cb(self, function):
        self._read_cb = function

    def read(self, chlst=None, params=None):
        if self._read_cb is not None:
            if self.hasChannels and chlst is not None:
                if params:
                    retValue = self._callbackChannels(self._read_cb, chlst,
                                                      params)
                else:
                    retValue = self._callbackChannels(self._read_cb, chlst)
            else:
                if params:
                    retValue = self._read_cb(params)
                else:
                    retValue = self._read_cb()
                self._debug("Attribute %s read: %s" % (self.name, retValue))
            return self._checkArray(retValue)

    def _checkArray(self, argin):
        # if answer is a list, manipulate it to follow the rule
        # '#NMMMMMMMMM...\n'
        isList = (type(argin) == list)
        isNpArray = None
        isSpArray = None
        if _np:
            if isList:
                argin = _np_ndarray(argin)
                isList = False
            isNpArray = (type(argin) == _np_ndarray)
        if _sp:
            isSpArray = (type(argin) == _sp_ndarray)
            if isList:
                argin = _sp_ndarray(argin)
                isList = False
        if isNpArray or isSpArray:
            argout = self._convertArray(argin)
            return argout
        return argin

    def _convertArray(self, argin):
        root = self._getRootComponent()
        dataFormat = root['dataFormat'].read()
        # flat the array, dimensions shall be known by the receiver
        flattened = argin.flatten()
        # codification
        if dataFormat == 'ASCII':
            data = "".join("%s," % element for element in flattened)[:-1]
        elif dataFormat == 'QUADRUPLE':
            data = flattened.astype(_float128).tostring()  # + '\n'
        elif dataFormat == 'DOUBLE':
            data = flattened.astype(_float64).tostring()  # + '\n'
        elif dataFormat == 'SINGLE':
            data = flattened.astype(_float32).tostring()  # + '\n'
        elif dataFormat == 'HALF':
            data = flattened.astype(_float16).tostring()  # + '\n'
        # TODO: mini-precision (byte)
        # elif dataFormat == 'mini':
        #     pass
        else:
            raise NotImplementedError("Unexpected data format %s codification"
                                      % (dataFormat))
        # prepare the header
        lenght = str(len(data))
        firstField = str(len(lenght))
        if len(firstField) > 1:
            self._error("A %s array cannot be codified" % (lenght))
            return float("NaN")
        header = "#%1s%s" % (firstField, lenght)
        return header + data

    def _getRootComponent(self):
        candidate = self._parent
        while candidate._parent is not None:
            candidate = candidate._parent
        return candidate

    @property
    def write_cb(self):
        return self._write_cb

    @write_cb.setter
    def write_cb(self, function):
        self._write_cb = function

    def write(self, chlst=None, value=None):
        self._debug("%s.write(ch=%s, value=%s)" % (self.name, chlst, value))
        if self._write_cb is not None:
            if self.allowedArgins is not None and \
                    value not in self.allowedArgins:
                raise ValueError("Not allowed to write %s, only %s "
                                 "are accepted" % (value, self.allowedArgins))
            if self.hasChannels and chlst is not None:
                retValue = self._callbackChannels(self._write_cb, chlst, value)
            else:
                retValue = self._write_cb(value)
                self._debug("Attribute %s write %s: %s"
                            % (self.name, value, retValue))
            return retValue

    def _callbackChannels(self, method_cb, chlst, value=None):
        self._checkAllChannelsAreWithinBoundaries(chlst)
        if len(chlst) == 1:
            ch = chlst[0]
            if value is None:
                retValue = method_cb(ch)
                self._debug("Attribute %s read from channel %d: %s"
                            % (self.name, ch, retValue))
            else:
                retValue = method_cb(ch, value)
                self._debug("Attribute %s write %s in channel %d: %s"
                            % (self.name, value, ch, retValue))
        else:
            if value is None:
                retValue = method_cb(chlst)
                self._debug("Attribute %s read for channel set %s: %s"
                            % (self.name, chlst, retValue))
            else:
                retValue = method_cb(chlst, value)
                self._debug("Attribute %s write %s for channel set %s: %s"
                            % (self.name, value, chlst, retValue))
        return retValue

    def _checkAllChannelsAreWithinBoundaries(self, chlst):
        if len(self._channelTree) != len(chlst):
            raise AssertionError("Given channel list hasn't the same number "
                                 "of elements than the known in the tree")
        for i, chRequested in enumerate(chlst):
            lowerBound = self._channelTree[i].firstChannel
            upperBound = lowerBound + self._channelTree[i].howManyChannels
            if chRequested < lowerBound:
                raise AssertionError("below the bounds")
            elif chRequested >= upperBound:
                raise AssertionError("above the bounds")


def BuildAttribute(name, parent, readcb=None, writecb=None, default=False,
                   allowedArgins=None):
    attr = Attribute(name)
    attr.parent = parent
    if parent is not None and name is not None:
        parent[name] = attr
    attr.read_cb = readcb
    attr.write_cb = writecb
    if default:
        parent.default = name
    if allowedArgins:
        attr.allowedArgins = allowedArgins
    attr.checkChannels()
    return attr


class Component(_Logger, dict):
    '''
        Intermediated nodes of the scpi command tree.

        Ex: COMPonent:COMPonent:ATTRibute
    '''
    def __init__(self, *args, **kargs):
        super(Component, self).__init__(*args, **kargs)
        self._parent = None
        self._defaultKey = None
        self._howMany = None
        self._hasChannels = False
        self._channelTree = None
        self._debug("Build a Component object %s" % (self.name))

    def __str__(self):
        fullName = "%s" % (self._name)
        parent = self._parent
        while parent is not None and parent._name is not None:
            fullName = "".join("%s:%s" % (parent._name, fullName))
            parent = parent._parent
        return fullName

    def __repr__(self):
        indentation = "\t"*self.depth
        repr = ""
        for key in self.keys():
            item = dict.__getitem__(self, key)
            # FIXME: ugly
            if isinstance(item, Attribute):
                repr = "".join("%s\n%s%r" % (repr, indentation, key))
            else:
                if item.default is not None:
                    isDefault = " (default %r) " % item.default
                else:
                    isDefault = ""
                if isinstance(item, Channel):
                    hasChannels = "NN"
                else:
                    hasChannels = ""
                repr = "".join("%s\n%s%r%s:%s%r"
                               % (repr, indentation, key, hasChannels,
                                  isDefault, item))
        return repr

    @property
    def name(self):
        return self._name

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        if value is not None:
            self.logEnable(value.logState())
            self.logLevel(value.logGetLevel())
            self._parent = value

    @property
    def default(self):
        return self._defaultKey

    @default.setter
    def default(self, value):
        if value in self.keys():
            self._defaultKey = value

    @property
    def hasChannels(self):
        return self._hasChannels

    # TODO: as any Attribute object will search for its channels in its parent
    #       component, as well as the components will search for channels also
    #       over their parents, and they are build from root to leafs:
    #       - once one finds a component with the channel feature, just copy
    #         its tree of the above channels already discovered.
    #       Right now this is following the same path over and over again.

    def checkChannels(self):
        if self.parent is not None and self.parent.hasChannels:
            self._hasChannels = True
            self._channelTree = self._getChannels()
            self._debug("%s: Channels found for %s component: %s"
                        % (self.name, self.parent.name,
                           ["%s" % x.name for x in self._channelTree]))
        else:
            self._debug("%s: No lower level channels found" % (self.name))

    def _getChannels(self):
        if self.parent is not None and self.parent.hasChannels:
            return self.parent._getChannels()
        return None

    def __getitem__(self, key):
        '''
            Given a keyword it checks if it matches, at least the first
            'minimumkey' characters, with an key in the dictionary to return
            its content.
        '''
        self._debug("available keywords: %s" % (self.keys()))
        for keyword in self.keys():
            if keyword == key:
                key = keyword
                break
        try:
            val = dict.__getitem__(self, key)
        except:
            raise KeyError("%s not found" % (key))
        self._debug("GET %s['%r'] = %s"
                    % (str(dict.get(self, 'name_label')), key, str(val)))
        return val

    def __setitem__(self, key, val):
        '''
            The key is case insensitive, then we store it as lower case to
            compare every where lower cases. Also the key corresponds to any
            substring of it with the minimum size of 'minimumKey'
        '''
        if type(key) != DictKey:
            key = DictKey(key)
        if not isinstance(val, (Component, Attribute)):
            raise ValueError("dictionary content shall be an attribute "
                             "or another Component (given %s)" % type(val))
        self._debug("SET %s['%r'] = %s"
                    % (str(dict.get(self, 'name_label')), key, str(val)))
        dict.__setitem__(self, key, val)
        val.parent = self

    def read(self, chlst=None, params=None):
        if self._defaultKey:
            return self.__getitem__(self._defaultKey).read(chlst, params)
        return float('NaN')

    def write(self, chlst=None, value=None):
        if self._defaultKey:
            return self.__getitem__(self._defaultKey).write(chlst, value)
        return float('NaN')


def BuildComponent(name=None, parent=None):
    component = Component(name=name)
    component.parent = parent
    if parent is not None and name is not None:
        parent[name] = component
    component.checkChannels()
    return component


class SpecialCommand(Component):
    '''
        Special commands that starts with '*' character. To be know, the
        mandatory commands that must have an instrument that provice scpi
        communications are:
        *IDN?: Identification query
                four field response:
                "MANUFACTURER,INSTRUMENT,SERIALNUMBER,FIRMWARE_VERSION
                    Manufacturer: identical for all the instruments of a
                                  single company
                    Instrument: It shall never contain the word 'MODEL'.
                    SN: specific for the responding instrument
                    version (and revision): of the software embedded in the
                                            instrument.
        *RST: Reset command
        *CLS: Clear status command
        *ESE[?]: Event Status Enable
        *ESR?: Event Status Register query
        *OPC[?]: Operation Complete
        *SRE[?]: Service Request Enable
        *STB?: Status Byte
        *TST?: self-test query
        *WAI: wait-to-continue command
    '''
    def __init__(self, *args, **kargs):
        super(SpecialCommand, self).__init__(*args, **kargs)
        self._readcb = None
        self._writecb = None

    @property
    def readcb(self):
        return self._readcb

    @readcb.setter
    def readcb(self, function):
        self._readcb = function

    def read(self):
        if self._readcb is not None:
            return self._readcb()
        return float("NaN")

    @property
    def writecb(self):
        return self._writecb

    @writecb.setter
    def writecb(self, function):
        self._writecb = function

    def write(self, value=None):
        if self._writecb:
            if value:
                self._writecb(value)
            else:
                self._writecb()
        return float("NaN")


def BuildSpecialCmd(name, parent, readcb, writecb=None):
    special = SpecialCommand(name)
    special.readcb = readcb
    special.writecb = writecb
    parent[name.lower()] = special
    return special


class Channel(Component):
    def __init__(self, howMany=None, startWith=1, *args, **kargs):
        super(Channel, self).__init__(*args, **kargs)
        if len(str(howMany).zfill(2)) > CHNUMSIZE:
            raise ValueError("The number of channels can not exceed "
                             "%d decimal digits" % (CHNUMSIZE))
        self._howMany = howMany
        self._startWith = startWith
        self._hasChannels = True
        self._channelTree = None
        self._debug("Build a Channel object %s" % (self.name))

    def _getChannels(self):
        if self.parent is not None and self.parent.hasChannels:
            parentChannels = self.parent._getChannels()
            if parentChannels is not None:
                return parentChannels + [self]
        return [self]

    @property
    def howManyChannels(self):
        return self._howMany

    @property
    def firstChannel(self):
        return self._startWith


def BuildChannel(name=None, howMany=None, parent=None, startWith=1):
    channel = Channel(name=name, howMany=howMany, startWith=startWith)
    channel.parent = parent
    if parent is not None and name is not None:
        parent[name] = channel
    channel.checkChannels()
    return channel



# -*- coding: utf-8 -*-
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

try:
    from .logger import Logger as _Logger
except:
    from logger import Logger as _Logger
from gc import collect as _gccollect
import socket as _socket
import threading as _threading
from time import sleep as _sleep
from traceback import print_exc as _print_exc

__author__ = "Sergi Blanch-Torné"
__email__ = "sblanch@cells.es"
__copyright__ = "Copyright 2015, CELLS / ALBA Synchrotron"
__license__ = "GPLv3+"


_MAX_CLIENTS = 10


class TcpListener(_Logger):
    """
        TODO: describe it
    """
    # FIXME: default should be local=False
    def __init__(self, name=None, callback=None, local=True, port=5025,
                 maxClients=_MAX_CLIENTS, ipv6=True, debug=False):
        super(TcpListener, self).__init__(debug=debug)
        self._name = name or "TcpListener"
        self._callback = callback
        self._connectionHooks = []
        self._local = local
        self._port = port
        self._maxClients = maxClients
        self._joinEvent = _threading.Event()
        self._joinEvent.clear()
        self._connectionThreads = {}
        self._withipv6suport = ipv6
        self.open()
        self._debug("Listener thread prepared")

    def __enter__(self):
        self._debug("received a enter() request")
        if not self.isOpen:
            self.open()
        return self

    def __exit__(self, type, value, traceback):
        self._debug("received a exit(%s,%s,%s) request"
                    % (type, value, traceback))
        self.__del__()

    def __del__(self):
        self.close()

    def open(self):
        self.buildIpv4Socket()
        try:
            self.buildIpv6Socket()
        except Exception as e:
            self._error("IPv6 will not be available due to: %s" % (e))

    def close(self):
        if self._joinEvent.isSet():
            return
        self._debug("%s close received" % (self._name))
        if hasattr(self, '_joinEvent'):
            self._debug("Deleting TcpListener")
            self._joinEvent.set()
        self._shutdownSocket(self._scpi_ipv4)
        if self._withipv6suport and hasattr(self, '_scpi_ipv6'):
            self._shutdownSocket(self._scpi_ipv6)
        if self._isListeningIpv4():
            self._scpi_ipv4 = None
        if self._withipv6suport and self._isListeningIpv6():
            self._scpi_ipv6 = None
        _gccollect()
        while self.isAlive():
            _gccollect()
            self._debug("Waiting for Listener threads")
            _sleep(1)
        self._debug("Everything is close, exiting...")

    @property
    def port(self):
        return self._port

    def listen(self):
        self._debug("Launching listener thread")
        self._listener_ipv4.start()
        if hasattr(self, '_listener_ipv6'):
            self._listener_ipv6.start()

    def isAlive(self):
        return self._isIPv4ListenerAlive() or self._isIPv6ListenerAlive()

    def isListening(self):
        return self._isListeningIpv4() or self._isListeningIpv6()

    def buildIpv4Socket(self):
        if self._local:
            self._host_ipv4 = '127.0.0.1'
        else:
            self._host_ipv4 = '0.0.0.0'
        self._scpi_ipv4 = _socket.socket(_socket.AF_INET, _socket.SOCK_STREAM)
        self._scpi_ipv4.setsockopt(_socket.SOL_SOCKET, _socket.SO_REUSEADDR, 1)
        self._listener_ipv4 = _threading.Thread(name="Listener4",
                                                target=self.__listener,
                                                args=(self._scpi_ipv4,
                                                      self._host_ipv4,))
        self._listener_ipv4.setDaemon(True)

    def buildIpv6Socket(self):
        if self._withipv6suport:
            if not _socket.has_ipv6:
                raise AssertionError("IPv6 not supported by the platform")
            if self._local:
                self._host_ipv6 = '::1'
            else:
                self._host_ipv6 = '::'
            self._scpi_ipv6 = _socket.socket(_socket.AF_INET6,
                                             _socket.SOCK_STREAM)
            self._scpi_ipv6.setsockopt(_socket.IPPROTO_IPV6,
                                       _socket.IPV6_V6ONLY, True)
            self._scpi_ipv6.setsockopt(_socket.SOL_SOCKET,
                                       _socket.SO_REUSEADDR, 1)
            self._listener_ipv6 = _threading.Thread(name="Listener6",
                                                    target=self.__listener,
                                                    args=(self._scpi_ipv6,
                                                          self._host_ipv6,))
            self._listener_ipv6.setDaemon(True)

    def _shutdownSocket(self, sock):
        try:
            sock.shutdown(_socket.SHUT_RDWR)
        except Exception as e:
            _print_exc()

    def _isIPv4ListenerAlive(self):
        return self._listener_ipv4.isAlive()

    def _isIPv6ListenerAlive(self):
        if hasattr(self, '_listener_ipv6'):
            return self._listener_ipv6.isAlive()
        return False

    def __listener(self, scpisocket, scpihost):
        try:
            self.__prepareListener(scpisocket, scpihost, 5)
            self.__doListen(scpisocket)
            self._debug("Listener thread finishing")
        except SystemExit as e:
            self._debug("Received a SystemExit (%s)" % (e))
            self.__del__()
        except KeyboardInterrupt as e:
            self._debug("Received a KeyboardInterrupt (%s)" % (e))
            self.__del__()
        except GeneratorExit as e:
            self._debug("Received a GeneratorExit (%s)" % (e))
            self.__del__()

    def __prepareListener(self, scpisocket, scpihost, maxretries):
        listening = False
        tries = 0
        seconds = 3
        while tries < maxretries:
            try:
                scpisocket.bind((scpihost, self._port))
                scpisocket.listen(self._maxClients)
                self._debug("Listener thread up and running (port %d, with "
                            "a maximum of %d connections in parallel)."
                            % (self._port, self._maxClients))
                return True
            except Exception as e:
                tries += 1
                self._error("Couldn't bind the socket. %s\nException: %s"
                            % ("(Retry in %d seconds)" % (seconds)
                               if tries < maxretries else "(No more retries)",
                               e))
                _sleep(seconds)
        return False

    def __doListen(self, scpisocket):
        while not self._joinEvent.isSet():
            try:
                connection, address = scpisocket.accept()
            except Exception as e:
                if self._joinEvent.isSet():
                    self._debug("Closing Listener")
                    del scpisocket
                    return
                self._error("Socket Accept Exception: %s" % (e))
                _sleep(3)
            else:
                self.__launchConnection(address, connection)
        scpisocket.close()

    def _isListeningIpv4(self):
        if hasattr(self, '_scpi_ipv4') and hasattr(self._scpi_ipv4, 'fileno'):
            return bool(self._scpi_ipv4.fileno())
        return False

    def _isListeningIpv6(self):
        if hasattr(self, '_scpi_ipv6') and hasattr(self._scpi_ipv6, 'fileno'):
            return bool(self._scpi_ipv6.fileno())
        return False

    @property
    def nActiveConnections(self):
        return len(self._connectionThreads)

    def __launchConnection(self, address, connection):
        connectionName = '%s:%s' % (address[0], address[1])
        try:
            self._debug('Connection request from %s (having %d already active)'
                        % (connectionName, self.nActiveConnections))
            if connectionName in self._connectionThreads and \
                    self._connectionThreads[connectionName].isAlive():
                self.error("New connection from %s when it has already "
                           "one. refusing the newer." % (connectionName))
            elif self.nActiveConnections >= self._maxClients:
                self._error("Reached the maximum number of allowed "
                            "connections (%d)" % (self.nActiveConnections))
            else:
                self._connectionThreads[connectionName] = \
                    _threading.Thread(name=connectionName,
                                      target=self.__connection,
                                      args=(address, connection))
                self._debug("Connection for %s created" % (connectionName))
                self._connectionThreads[connectionName].setDaemon(True)
                self._connectionThreads[connectionName].start()
        except Exception as e:
            self._error("Cannot launch connection request from %s due to: %s"
                        % (connectionName, e))

    def __connection(self, address, connection):
        connectionName = '%s:%s' % (address[0], address[1])
        self._debug("Thread for %s connection" % (connectionName))
        while not self._joinEvent.isSet():
            data = connection.recv(1024)
            self._info("received from %s: %d bytes %r"
                       % (connectionName, len(data), data))
            if len(self._connectionHooks) > 0:
                for hook in self._connectionHooks:
                    try:
                        hook(connectionName, data)
                    except Exception as e:
                        self._warning("Exception calling %s hook: %s"
                                      % (hook, e))
            if len(data) == 0:
                self._warning("No data received, termination the connection")
                connection.close()
                break
            if self._callback is not None:
                ans = self._callback(data)
                self._debug("scpi.input say %r" % (ans))
                connection.send(ans)
        self._connectionThreads.pop(connectionName)
        self._debug("Ending connection: %s (having %s active left)"
                    % (connectionName, self.nActiveConnections))

    def addConnectionHook(self, hook):
        if callable(hook):
            self._connectionHooks.append(hook)
        else:
            raise TypeError("The hook must be a callable object")

    def removeConnectionHook(self, hook):
        if self._connectionHooks.count(hook):
            self._connectionHooks.pop(self._connectionHooks.index(hook))
            return True
        return False



# -*- coding: utf-8 -*-
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Sergi Blanch-Torne"
__email__ = "sblanch@cells.es"
__copyright__ = "Copyright 2015, CELLS / ALBA Synchrotron"
__license__ = "GPLv3+"

__version__ = '0.4.0'


def VERSION():
    if '-' in __version__:
        _v, _rel = __version__.split('-')
    else:
        _v, _rel = __version__, ''
    _v = [int(n) for n in _v.split('.')]
    if len(_rel) > 0:
        _v += [_rel]
    return tuple(_v)


def version():
    return __version__



# -*- coding: utf-8 -*-
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####


from datetime import datetime as _datetime
from datetime import timedelta as _timedelta
try:
    from .logger import Logger as _Logger
except:
    from logger import Logger as _Logger
from threading import currentThread as _currentThread


__author__ = "Sergi Blanch-Torné"
__copyright__ = "Copyright 2016, CELLS / ALBA Synchrotron"
__license__ = "GPLv3+"

__all__ = ["Locker"]


DEFAULT_EXPIRATION_TIME = 60  # seconds
UPPER_LIMIT_EXPIRATION_TIME = _timedelta(0, DEFAULT_EXPIRATION_TIME*10, 0)


class Locker(_Logger):
    """
        Object to control the access to certain areas of the code, similar idea
        than a Semaphore or a Lock from the threading library, but not the
        same.

        An external process talks with a service thread, and what the thread
        does
    """
    def __init__(self, *args, **kargs):
        super(Locker, self).__init__(*args, **kargs)
        self._owner = None
        self._when = None
        self._expiration = _timedelta(0, DEFAULT_EXPIRATION_TIME, 0)

    def __str__(self):
        return "%s(%s)"\
            % (self.name, "" if self._owner is None else self._owner)

    def __repr__(self):
        return "%s(owner=%s, when=%s, expiration=%s)"\
            % (self.name, self._owner, self._when, self._expiration)

    @property
    def owner(self):
        return self._owner

    def Owner(self):
        return self.owner

    # API ---

    def request(self, timeout=None):
        """
            Request to book the access. If its free (or has expired a previous
            one) do it, else reject (even if the owner recalls it).
        """
        # self._debug("%s request()" % _currentThread().name)
        if not self._hasOwner() or self._hasExpired():
            self._doLock(timeout)
            return True
        else:
            self._warning("%s request the lock when already is"
                          % (_currentThread().name))
            return False

    def release(self):
        """
            Only the owner can release the lock.
        """
        # self._debug("%s release()" % _currentThread().name)
        if self._isOwner():
            self._debug("%s releases the lock" % self._owner)
            self._doRelease()
            return True
        else:
            self._error("%s is NOT allowed to release %s's lock"
                        % (_currentThread().name, self._owner))
            return False

    def access(self):
        """
            Request if the give thread is allowed to access the resource.
        """
        # self._debug("%s access()" % _currentThread().name)
        if not self.isLock():
            self._debug("There is no owner of the lock, %s can pass"
                        % _currentThread().name)
            return True
        elif self._isOwner():
            self._when = _datetime.now()
            self._debug("The owner is who is asking (%s), "
                        "renew its booking" % (_currentThread().name))
            return True
        else:
            return False

    def isLock(self):
        """
            Check if the lock is owned.
        """
        # self._debug("%s isLock()" % _currentThread().name)
        if self._hasOwner() and not self._hasExpired():
            return True
        return False

    @property
    def expirationTime(self):
        return self._expiration

    @expirationTime.setter
    def expirationTime(self, value):
        if not self._hasExpired() or self._isOwner():
            self._setExpiration(value)
        else:
            raise EnvironmentError("Only the owner can change the expiration")

    # force area ---

    def _forceRelease(self):
        self._warning("Locker forced to be released!")
        self._doRelease()

    def _forceLock(self):
        self._forceRelease()
        self._doLock()

    # internal methods ---

    def _hasOwner(self):
        hasOwner = self._owner is not None
#         if hasOwner:
#             self._debug("Lock has owner (%s)" % self._owner)
#         else:
#             self._debug("Lock is free")
        return hasOwner

    def _isOwner(self):
        isOwner = self._owner == _currentThread().name
#         if isOwner:
#             self._debug("%s is the owner" % (_currentThread().name))
#         else:
#             self._debug("%s is NOT the owner" % (_currentThread().name))
        return isOwner

    def _doLock(self, timeout=None):
        if self._hasOwner():
            raise RuntimeError("Try to lock when not yet released")
        self._owner = _currentThread().name
        self._when = _datetime.now()
        self.expirationTime = timeout
        self._info("%s has take the lock (expiration time %s)"
                   % (self.owner, self.expirationTime))

    def _doRelease(self):
        self._owner = None
        self._when = None
        self._expiration = None

    def _setExpiration(self, value):
        if value is None:
            value = DEFAULT_EXPIRATION_TIME
        try:
            value = int(value)
        except Exception as e:
            raise TypeError("Expiration time assignment must be an integer")
        try:
            value = _timedelta(0, value, 0)
        except Exception as e:
            raise ValueError("Value cannot be understood as a delta time")
        if value > UPPER_LIMIT_EXPIRATION_TIME:
            raise OverflowError("Too big expiration time")
        else:
            self._expiration = value

    def _hasExpired(self):
        if self._when is None or self._expiration is None:
            # TBD: in fact here there is nothing to expire
            return True
        delta = _datetime.now() - self._when
        if delta > self._expiration:
            self._info("No news from the lock owner (%s) after %s: release "
                       "the lock." % (self._owner, delta))
            self._doRelease()
            return True
        self._debug("lock NOT expired, still %s for %s" % (delta, self._owner))
        return False



# -*- coding: utf-8 -*-
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####


try:
    from .commands import Component, Attribute, BuildComponent, BuildChannel
    from .commands import BuildAttribute, BuildSpecialCmd, CHNUMSIZE
    from .logger import Logger as _Logger
    from .tcpListener import TcpListener
    from .lock import Locker as _Locker
    from .version import version as _version
except:
    from commands import Component, Attribute, BuildComponent, BuildChannel
    from commands import BuildAttribute, BuildSpecialCmd, CHNUMSIZE
    from logger import Logger as _Logger
    from tcpListener import TcpListener
    from lock import Locker as _Locker
    from version import version as _version
from time import sleep as _sleep
from time import time as _time
from threading import currentThread as _currentThread
from traceback import print_exc


__author__ = "Sergi Blanch-Torné"
__copyright__ = "Copyright 2015, CELLS / ALBA Synchrotron"
__license__ = "GPLv3+"

__all__ = ["scpi"]


# DEPRECATED: flags for service activation
TCPLISTENER_LOCAL = 0b10000000
TCPLISTENER_REMOTE = 0b01000000


def __version__():
    '''Library version with 4 fields: 'a.b.c-d'
       Where the two first 'a' and 'b' comes from the base C library
       (scpi-parser), the third is the build of this cypthon port and the last
       one is a revision number.
    '''
    return _version()


class scpi(_Logger):
    '''This is an object to be build in order to provide to your instrument
       SCPI communications. By now it only provides network (ipv4 and ipv6)
       communications, but if required it can be extended to support other
       types.

       By default it builds sockets that only listen in the loopback network
       interface. By default we like to avoid to expose the conection. There
       are two ways to allow direct remote connections. One in the constructor
       by calling it with the parameter services=TCPLISTENER_REMOTE. The other
       can be called once the object is created by setting the object property
       'remoteAllowed' to True.
    '''
    def __init__(self, commandTree=None, specialCommands=None,
                 local=True, port=5025, autoOpen=False,
                 services=None, writeLock=False, *args, **kwargs):
        super(scpi, self).__init__(*args, **kwargs)
        self._name = "scpi"
        self._commandTree = commandTree or Component()
        self._commandTree.logEnable(self.logState())
        self._commandTree.logLevel(self.logGetLevel())
        self._specialCmds = specialCommands or {}
        self._debug("Special commands: %r" % (specialCommands))
        self._debug("Given commands: %r" % (self._commandTree))
        self._local = local
        self._port = port
        self._services = {}
        if services is not None:
            msg = "The argument 'services' is deprecated, "\
                  "please use the boolean 'local'"
            header = "*"*len(msg)
            if services & (TCPLISTENER_LOCAL | TCPLISTENER_REMOTE):
                self._local = bool(services & TCPLISTENER_LOCAL)
        if autoOpen:
            self.open()
        self.__buildDataFormatAttribute()
        self.__buildSystemComponent(writeLock)

    def __enter__(self):
        self._debug("received a enter() request")
        if not self.isOpen:
            self.open()
        return self

    def __exit__(self, type, value, traceback):
        self._debug("received a exit(%s,%s,%s) request"
                    % (type, value, traceback))
        if self.isOpen:
            self.close()

    def __del__(self):
        self._debug("Delete request received")
        if self.isOpen:
            self.close()

    def __str__(self):
        return "%s" % (self.name)

    def __repr__(self):
        if 'idn' in self._specialCmds:
            return "%s(%s)" % (self.name, self._specialCmds['idn'].read())
        return "%s()" % self.name

    # # communications ares ---

    # TODO: status information method ---
    #       (report the open listeners and accepted connections ongoing).
    # TODO: other incoming channels than network ---

    @property
    def isOpen(self):
        return any([self._services[key].isListening()
                    for key in self._services.keys()])

    def open(self):
        if not self.isOpen:
            self.__buildTcpListener()
        else:
            self._warning("Already Open")

    def close(self):
        if self.isOpen:
            self._debug("Close services")
            for key in self._services.keys():
                self._debug("Close service %s" % key)
                self._services[key].close()
                self._services.pop(key)
            self._debug("Communications finished. Exiting...")
        else:
            self._warning("Already Close")

    def __buildTcpListener(self):
        self._debug("Opening tcp listener (%s)"
                    % ("local" if self._local else "remote"))
        self._services['tcpListener'] = TcpListener(name="TcpListener",
                                                    callback=self.input,
                                                    local=self._local,
                                                    port=self._port,
                                                    debug=self.logState())
        self._services['tcpListener'].listen()

    def addConnectionHook(self, hook):
        for service in self._services.itervalues():
            if hasattr(service, 'addConnectionHook'):
                try:
                    service.addConnectionHook(hook)
                except Exception as e:
                    self._error("Exception setting a hook to %s: %s"
                                % (service, e))
            else:
                self._warning("Service %s doesn't support hooks" % (service))

    def removeConnectionHook(self, hook):
        for service in self._services.itervalues():
            if hasattr(service, 'removeConnectionHook'):
                if not service.removeConnectionHook(hook):
                    self._warning("Service %s refuse to remove the hook"
                                  % (service))
            else:
                self._warning("Service %s doesn't support hooks" % (service))

    def __buildDataFormatAttribute(self):
        self._dataFormat = 'ASCII'
        self.addAttribute('DataFormat', self._commandTree,
                          self.dataFormat, self.dataFormat,
                          allowedArgins=['ASCII', 'QUADRUPLE', 'DOUBLE',
                                         'SINGLE', 'HALF'])

    def __buildSystemComponent(self, writeLock):
        systemTree = self.addComponent('system', self._commandTree)
        self.__buildLockerComponent(systemTree)
        if writeLock:
            self.__buildWLockerComponent(systemTree)
        else:
            self._wlock = None
        # TODO: other SYSTem components from SCPI-99 (pages 21-*)
        #       :SYSTem:PASSword
        #       :SYSTem:PASSword:CDISable
        #       :SYSTem:PASSword:CENAble
        #       :SYSTem:PASSword:NEW
        #       :SYSTem:PRESet
        #       :SYSTem:SECUrity
        #       :SYSTem:SECUrity:IMMEdiate
        #       :SYSTem:SECUrity:STATe
        #       :SYSTem:TIME
        #       :SYSTem:TIME:TIMEr
        #       :SYSTem:TIME:TIMEr:COUNt
        #       :SYSTem:TIME:TIMEr:STATe
        #       :SYSTem:TZONe
        #       :SYSTem:VERSion

    def __buildLockerComponent(self, commandTree):
        self._lock = _Locker(name='readLock', debug=self.logState())
        subTree = self.addComponent('LOCK', commandTree)
        self.addAttribute('owner', subTree, self._lock.Owner, default=True)
        self.addAttribute('release', subTree, readcb=self._lock.release,
                          writecb=self._lock.release)
        self.addAttribute('request', subTree,
                          readcb=self._lock.request,
                          writecb=self._lock.request)

    def __buildWLockerComponent(self, commandTree):
        self._wlock = _Locker(name='writeLock', debug=self.logState())
        subTree = self.addComponent('WLOCK', commandTree)
        self.addAttribute('owner', subTree, self._wlock.Owner, default=True)
        self.addAttribute('release', subTree, readcb=self._wlock.release,
                          writecb=self._wlock.release)
        self.addAttribute('request', subTree,
                          readcb=self._wlock.request,
                          writecb=self._wlock.request)

    @property
    def remoteAllowed(self):
        return not self._services['tcpListener']._local

    @remoteAllowed.setter
    def remoteAllowed(self, value):
        if type(value) is not bool:
            raise AssertionError("Only boolean can be assigned")
        if value != (not self._services['tcpListener']._local):
            tcpListener = self._services.pop('tcpListener')
            tcpListener.close()
            self._debug("Close the active listeners and their connections.")
            while tcpListener.isListening():
                self._warning("Waiting for listerners finish")
                _sleep(1)
            self._debug("Building the new listeners.")
            if value is True:
                self.__buildTcpListener(TCPLISTENER_REMOTE)
            else:
                self.__buildTcpListener(TCPLISTENER_LOCAL)
        else:
            self._debug("Nothing to do when setting like it was.")

    # done communications area ---

    # # command introduction area ---

    def addSpecialCommand(self, name, readcb, writecb=None):
        '''
            Adds a command '*%s'%(name). If finishes with a '?' mark it will
            be called the readcb method, else will be the writecb method.
        '''
        name = name.lower()
        if name.startswith('*'):
            name = name[1:]
        if name.endswith('?'):
            if writecb is not None:
                raise KeyError("Refusing command %s: looks readonly but has "
                               "a query character at the end." % (name))
            name = name[:-1]
        if not name.isalpha():
            raise NameError("Not supported other than alphabetical characters")
        if self._specialCmds is None:
            self._specialCmds = {}
        self._debug("Adding special command '*%s'" % (name))
        BuildSpecialCmd(name, self._specialCmds, readcb, writecb)

    @property
    def specialCommands(self):
        return self._specialCmds.keys()

    def addComponent(self, name, parent):
        if not hasattr(parent, 'keys'):
            raise TypeError("For %s, parent doesn't accept components"
                            % (name))
        if name in parent.keys():
            self._debug("component '%s' already exist" % (name))
            return
        self._debug("Adding component '%s' (%s)" % (name, parent))
        return BuildComponent(name, parent)

    def addChannel(self, name, howMany, parent, startWith=1):
        if not hasattr(parent, 'keys'):
            raise TypeError("For %s, parent doesn't accept components"
                            % (name))
        if name in parent.keys():
            self._debug("component '%s' already exist" % (name))
            return
        self._debug("Adding component '%s' (%s)" % (name, parent))
        return BuildChannel(name, howMany, parent, startWith)

    def addAttribute(self, name, parent, readcb, writecb=None, default=False,
                     allowedArgins=None):
        if not hasattr(parent, 'keys'):
            raise TypeError("For %s, parent doesn't accept attributes"
                            % (name))
        if name in parent.keys():
            self._debug("attribute '%s' already exist" % (name))
            return
        self._debug("Adding attribute '%s' (%s)" % (name, parent))
        return BuildAttribute(name, parent, readcb, writecb, default,
                              allowedArgins)

    def addCommand(self, FullName, readcb, writecb=None, default=False,
                   allowedArgins=None):
        '''
            adds the command in the structure of [X:Y:]Z composed by Components
            X, Y and as many as ':' separated have. The last one will
            correspond with an Attribute with at least a readcb for when it's
            called with a '?' at the end. Or writecb if it's followed by an
            space and something that can be casted after.
        '''
        if FullName.startswith('*'):
            self.addSpecialCommand(FullName, readcb, writecb)
            return
        nameParts = FullName.split(':')
        self._debug("Prepare to add command %s" % (FullName))
        tree = self._commandTree
        # preprocessing:
        for i, part in enumerate(nameParts):
            if len(part) == 0:
                raise NameError("No null names allowed "
                                "(review element %d of %s)" % (i, FullName))
        if len(nameParts) > 1:
            for i, part in enumerate(nameParts[:-1]):
                self.addComponent(part, tree)
                tree = tree[part]
        self.addAttribute(nameParts[-1], tree, readcb, writecb, default,
                          allowedArgins)

    # done command introduction area ---

    # # input/output area ---

    @property
    def commands(self):
        return self._commandTree.keys()

    def dataFormat(self, value=None):
        if value is None:
            return self._dataFormat
        self._dataFormat = value

    def input(self, line):
        self._debug("Received %r input" % (line))
#         if not self._isAccessAllowed():
#             return ''
        start_t = _time()
        while len(line) > 0 and line[-1] in ['\r', '\n', ';']:
            self._debug("from %r remove %r" % (line, line[-1]))
            line = line[:-1]
        if len(line) == 0:
            return ''
        line = line.split(';')
        results = []
        for i, command in enumerate(line):
            command = command.strip()  # avoid '\n' terminator if exist
            self._debug("Processing %dth command: %r" % (i+1, command))
            if command.startswith('*'):
                answer = self._process_special_command(command[1:])
                if answer is not None:
                    results.append(answer)
            elif command.startswith(':'):
                if i == 0:
                    self._error("For command %r: Not possible to start "
                                "with ':', without previous command"
                                % (command))
                    results.append(float('NaN'))
                else:
                    # populate fields pre-':'
                    # with the previous (i-1) command
                    command = "".join("%s%s" % (line[i-1].rsplit(':', 1)[0],
                                                command))
                    self._debug("Command expanded to %r" % (command))
                    answer = self._process_normal_command(command)
                    if answer is not None:
                        results.append(answer)
            else:
                answer = self._process_normal_command(command)
                if answer is not None:
                    results.append(answer)
        self._debug("Answers: %r" % (results))
        answer = ""
        for res in results:
            answer = "".join("%s%s;" % (answer, res))
        self._debug("Answer: %r" % (answer))
        self._debug("Query reply send after %g ms" % ((_time()-start_t)*1000))
        # FIXME: has the last character to be ';'?
        if len(answer[:-1]):
            return answer[:-1]+'\r\n'
        # return answer + '\r\n'
        return ''

    def _process_special_command(self, cmd):
        start_t = _time()
        result = None
        # FIXME: ugly
        self._debug("current special keys: %s" % (self._specialCmds.keys()))
        if cmd.count(':') > 0:  # Not expected in special commands
            return float('NaN')
        for key in self._specialCmds.keys():
            self._debug("testing key %s ?= %s" % (key, cmd))
            if cmd.lower().startswith(key.lower()):
                if cmd.endswith('?'):
                    if self._isAccessAllowed():
                        self._debug("Requesting read of %s" % (key))
                        result = self._specialCmds[key].read()
                        break
                    else:
                        result = float('NaN')
                        break
                if cmd.count(' ') > 0:
                    if self._isAccessAllowed() and\
                            self._isWriteAccessAllowed():
                        bar = cmd.split(' ')
                        name, value = bar
                        self._debug("Requesting write of %s with value %s"
                                    % (name, value))
                        # TODO: By default don't provide a readback,
                        #       but there will be an SCPI command to return
                        #       an answer to the write commands
                        self._specialCmds[name].write(value)
                    break
                self._debug("Requesting write of %s without value" % (key))
                # TODO: By default don't provide a readback,
                #       but there will be an SCPI command to return
                #       an answer to the write commands
                self._specialCmds[key].write()
                break
        self._debug("special command %s processed in %g ms"
                    % (cmd, (_time()-start_t)*1000))
        return result

    def _process_normal_command(self, cmd):
        start_t = _time()
        answer = None
        keywords = cmd.split(':')
        tree = self._commandTree
        channelNum = []
        for key in keywords:
            self._debug("processing %s" % key)
            key, separator, params = self._splitParams(key)
            key = self._check4Channels(key, channelNum)
            try:
                nextNode = tree[key]
                if separator == '?':
                    if self._isAccessAllowed():
                        answer = self._doReadOperation(cmd, tree, key,
                                                       channelNum, params)
                elif separator == ' ' or type(nextNode) == Attribute:
                    # with separator next comes the parameters, without it is
                    # a (write) command without parameters. But in this second
                    # case it must by an Attribute component or it may confuse
                    # with intermediate keys of the command.
                    if self._isAccessAllowed() and \
                            self._isWriteAccessAllowed():
                        self._doWriteOperation(cmd, tree, key, channelNum,
                                               params)
                else:
                    tree = nextNode
            except Exception as e:
                self._error("Not possible to understand key %r (from %r) "
                            "separator %r, params %r" % (key, cmd, separator,
                                                         params))
                if separator == '?':
                    answer = float('NaN')
                print_exc()
                break
        self._debug("command %s processed in %g ms (%r)"
                    % (cmd, (_time()-start_t)*1000, answer))
        return answer

    def _splitParams(self, key):
        for separator, operation in [['?', 'read'], [' ', 'write']]:
            if key.count(separator):
                key, params = key.split(separator)
                if len(params) > 0:
                    self._debug("Found a %s with params: key=%s, params=%s"
                                % (operation, key, params))
                    return key, separator, params
                return key, separator, None
        return key, None, None

    def _check4Channels(self, key, channelNum):
        if key[-CHNUMSIZE:].isdigit():
            channelNum.append(int(key[-CHNUMSIZE:]))
            self._debug("It has been found that this has channels defined "
                        "for keyword %s" % (key))
            key = key[:-CHNUMSIZE]
        return key

    def _doReadOperation(self, cmd, tree, key, channelNum, params):
        try:
            self._debug("Leaf of the tree %r%s"
                        % (key, " (with params=%s)"
                           % params if params else ""))
            if len(channelNum) > 0:
                self._debug("do read with channel")
                if params:
                    answer = tree[key].read(chlst=channelNum,
                                            params=params)
                else:
                    answer = tree[key].read(chlst=channelNum)
            else:
                if params:
                    answer = tree[key].read(params=params)
                else:
                    answer = tree[key].read()
            # With the support for list readings (its conversion
            # to '#NMMMMMMMMM...' stream:
            # TODO: This will require a DataFormat feature to
            #       pack the data in bytes, shorts or longs.
            if answer is None:
                answer = float('NaN')
        except Exception as e:
            self._warning("Exception reading '%s': %s" % (cmd, e))
            answer = float('NaN')
            print_exc()
        return answer

    def _doWriteOperation(self, cmd, tree, key, channelNum, params):
        # TODO: By default don't provide a readback, but there will be an SCPI
        #       command to return an answer to the write commands
        try:
            self._debug("Leaf of the tree %r (%r)" % (key, params))
            if len(channelNum) > 0:
                self._debug("do write (with channel %s) %s: %s"
                            % (channelNum, key, params))
                answer = tree[key].write(chlst=channelNum, value=params)
#                 if answer is None:
#                     answer = tree[key].read(channelNum)
            else:
                self._debug("do write %s: %s" % (key, params))
                answer = tree[key].write(value=params)
#                 if answer is None:
#                     answer = tree[key].read()
        except Exception as e:
            self._warning("Exception writing '%s': %s" % (cmd, e))
            answer = None  # float('NaN')
            print_exc()
        return answer

    # input/output area ---

    # # lock access area ---
    def _BookAccess(self):
        return self._lock.request()

    def _UnbookAccess(self):
        return self._lock.release()

    def _isAccessAllowed(self):
        return self._lock.access()

    def _isAccessBooked(self):
        return self._lock.isLock()

    def _forceAccessRelease(self):
        self._lock._forceRelease()

    def _forceAccessBook(self):
        self._forceAccessRelease()
        return self._BookAccess()

    def _LockOwner(self):
        return self._lock.owner

    def _BookWriteAccess(self):
        if self._wlock:
            return self._wlock.request()
        return False

    def _UnbookWriteAccess(self):
        if self._wlock:
            return self._wlock.release()
        return False

    def _isWriteAccessAllowed(self):
        if self._wlock:
            return self._wlock.access()
        return True

    def _isWriteAccessBooked(self):
        if self._wlock:
            return self._wlock.isLock()
        return False

    def _forceWriteAccessRelease(self):
        if self._wlock:
            self._wlock._forceRelease()

    def _forceWriteAccessBook(self):
        self._forceAccessRelease()
        return self._BookAccess()

    def _WLockOwner(self):
        if self._wlock:
            return self._wlock.owner
        return None
    # lock access area ---



# -*- coding: utf-8 -*-
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

from setuptools import setup, find_packages

__author__ = "Sergi Blanch-Torne"
__email__ = "sblanch@cells.es"
__copyright__ = "Copyright 2015, CELLS / ALBA Synchrotron"
__license__ = "GPLv3+"

__project__ = 'scpilib'
__description__ = "Python module to provide scpi functionality "\
                  "from instrument side"
__longDesc__ = """
This module has been prepared to provide an instrument the functionality
to be accessed via the SCPI Protocol.

By now it only supports network connection, by default port 5025.
"""
__url__ = "https://github.com/srgblnch/python-scpilib"
# we use semantic versioning (http://semver.org/) and we update it using the
# bumpversion script (https://github.com/peritus/bumpversion)
__version__ = '0.4.0'


setup(name=__project__,
      license=__license__,
      description=__description__,
      long_description=__longDesc__,
      version=__version__,
      author=__author__,
      author_email=__email__,
      classifiers=['Development Status :: 2 - Pre-Alpha',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: '
                   'GNU General Public License v3 or later (GPLv3+)',
                   'Operating System :: POSIX',
                   'Programming Language :: Python',
                   'Topic :: Scientific/Engineering :: '
                   'Interface Engine/Protocol Translator',
                   'Topic :: Software Development :: Embedded Systems',
                   'Topic :: Software Development :: Libraries :: '
                   'Python Modules',
                   'Topic :: System :: Hardware',
                   ],
      packages=find_packages(),
      url=__url__,
      )

# for the classifiers review see:
# https://pypi.python.org/pypi?%3Aaction=list_classifiers
#
# Development Status :: 1 - Planning
# Development Status :: 2 - Pre-Alpha
# Development Status :: 3 - Alpha
# Development Status :: 4 - Beta
# Development Status :: 5 - Production/Stable



# -*- coding: utf-8 -*-
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

from datetime import datetime as _datetime
import logging as _logging
from logging import handlers as _handlers
from multiprocessing import current_process as _currentProcess
import os
from threading import currentThread as _currentThread
from threading import Lock as _Lock
from weakref import ref as _weakref


__author__ = "Sergi Blanch-Torné"
__email__ = "sblanch@cells.es"
__copyright__ = "Copyright 2015, CELLS / ALBA Synchrotron"
__license__ = "GPLv3+"


global lock
lock = _Lock()

_logger_NOTSET = _logging.NOTSET  # 0
_logger_CRITICAL = _logging.CRITICAL  # 50
_logger_ERROR = _logging.ERROR  # 40
_logger_WARNING = _logging.WARNING  # 30
_logger_INFO = _logging.INFO  # 20
_logger_DEBUG = _logging.DEBUG  # 10


__all__ = ["Logger"]


class Logger(object):
    '''This class is a very basic debugging flag mode used as a super class
       for the other classes in this library.
    '''

    _levelStr = {_logger_NOTSET:   '',
                 _logger_CRITICAL: 'CRITICAL',
                 _logger_ERROR:    'ERROR',
                 _logger_WARNING:  'WARNING',
                 _logger_INFO:     'INFO',
                 _logger_DEBUG:    'DEBUG'}

    def __init__(self, name="Logger", debug=False, loggerName=None,
                 log2File=False, *args, **kwargs):
        super(Logger, self).__init__(*args, **kwargs)
        # prepare vbles ---
        self._name = name
        self.__debugFlag = None
        self.__debuglevel = _logger_NOTSET
        self.__log2file = log2File
        self.__loggerName = loggerName or "SCPI"
        self.__logging_folder = None
        self.__logging_file = None
        self._devlogger = _logging.getLogger(self.__loggerName)
        self._handler = None
        # setup ---
        self.logEnable(True)
        if debug:
            self.logLevel(_logger_DEBUG)
        else:
            self.logLevel(_logger_INFO)
        self.loggingFile()
        if not len(self._devlogger.handlers):
            self._devlogger.setLevel(_logger_DEBUG)
            self._handler = \
                _handlers.RotatingFileHandler(self.__logging_file,
                                              maxBytes=10000000,
                                              backupCount=5)
            self._handler.setLevel(_logger_NOTSET)
            formatter = _logging.Formatter('%(asctime)s - %(levelname)s - '
                                           '%(name)s - %(message)s')
            self._handler.setFormatter(formatter)
            self._devlogger.addHandler(self._handler)
        else:
            self._handler = self._devlogger.handlers[0]

    @property
    def name(self):
        return self._name

    @property
    def depth(self):
        depth = 0
        if hasattr(self, '_parent'):
            parent = self._parent
            while parent is not None:
                parent = parent._parent
                depth += 1
        return depth

    @property
    def devlogger(self):
        return self._devlogger

    @devlogger.setter
    def devlogger(self, devlogger):
        self._devlogger = devlogger

    @property
    def handler(self):
        return self._handler

    def addHandler(self, handler):
        self._devlogger.addHandler(handler)

    def removeHandler(self, handler=None):
        if handler:
            self._devlogger.removeHandler(handler)
        else:
            self._devlogger.removeHandler(self._handler)

    def replaceHandler(self, handler):
        self._devlogger.removeHandler(self._handler)
        self._devlogger.addHandler(handler)
        self._handler = handler

    def loggingFolder(self):
        if self.__logging_folder is None:
            logging_folder = "/var/log/%s" % (self.__loggerName)
            if not self.__buildLoggingFolder(logging_folder):
                logging_folder = "/tmp/log/%s" % (self.__loggerName)
                if not self.__buildLoggingFolder(logging_folder):
                    raise SystemError("No folder for logging available")
            self.__logging_folder = logging_folder
        else:
            if not self.__buildLoggingFolder(self.__logging_folder):
                raise SystemError("No folder for logging available")
        return self.__logging_folder

    def __buildLoggingFolder(self, folder):
        try:
            if not os.path.exists(folder):
                os.makedirs(folder)
            return True
        except:
            return False

    def loggingFile(self):
        if self.__logging_file is None:
            self.__logging_file = \
                "%s/%s.log" % (self.loggingFolder(), self.__loggerName)
        return self.__logging_file

    def log2File(self, boolean):
        if type(boolean) is not bool:
            raise AssertionError("The parameter must be a boolean")
        self.__log2file = boolean
        self.logMessage("log2File() set to %s"
                        % (self.__log2file), _logger_INFO)

    def logEnable(self, dbg=False):
        if type(dbg) is not bool:
            raise AssertionError("The parameter must be a boolean")
        self.__debugFlag = dbg
        self.logMessage("logEnable()::Debug flag set to %s"
                        % (self.__debugFlag), _logger_INFO)

    def logState(self):
        return self.__debugFlag

    def logLevel(self, level):
        try:
            self.__debuglevel = int(level)
        except:
            raise AssertionError("The loglevel must be an integer")
        self._devlogger.setLevel(level)
        if self._handler is not None:
            self._handler.setLevel(level)
        self.logMessage("logEnable()::Debug level set to %s"
                        % (self.__debuglevel), _logger_INFO)
#         if self._handler is not None:
#             self._handler.setLevel(level)
#         self.logMessage("logEnable()::Debug level set to %s"
#                         % (self.__debuglevel), _logger_INFO)

    def logGetLevel(self):
        return self.__debuglevel

    @property
    def _processId(self):
        return _currentProcess().name

    @property
    def _threadId(self):
        return _currentThread().getName()

    def logMessage(self, msg, level):
        _tag = self._levelStr[level]
        prt_msg = "%s - %s - %s" % (_tag, self.__loggerName, msg)
        if self.__log2file:
            method = {_logger_CRITICAL: self._devlogger.critical,
                      _logger_ERROR: self._devlogger.error,
                      _logger_WARNING: self._devlogger.warn,
                      _logger_INFO: self._devlogger.info,
                      _logger_DEBUG: self._devlogger.debug}
            method[level](msg)

    def _critical(self, msg):
        self.logMessage(msg, _logger_CRITICAL)

    def _error(self, msg):
        self.logMessage(msg, _logger_ERROR)

    def _warning(self, msg):
        self.logMessage(msg, _logger_WARNING)

    def _info(self, msg):
        self.logMessage(msg, _logger_INFO)

    def _debug(self, msg):
        if self.__debugFlag:
            self.logMessage(msg, _logger_DEBUG)
